/*
    PROGRAM NAME: ROTO-MICROSCOPE CONTROLLER

    DESCRIPTION : This is the program written for the Arduino Mega to control the Mighty Roto_microscope.
                  The joystick module is used to rotate  the microscope (y-axis) and the microscope table (x-axis).
                  The button is used to take pictures using the mouse-shutter.

                  KEYPAD CONTROL
                  up/down    : LCD menu item selection
                  Left/Right : Change the values of the the rotation angles.
                  Right      : Activate the automatic sequence (Menu item: Auto. Mode).  
                  Select:    : Move the microscope down to the home position and in automatic mode mode to abort.

    LCD MENU    :
                  X-steps/degree:  Set up the rotation steps (in degrees) of the microscope rotating table.
                  Y-steps/degrees: Set up the rotation steps (in degrees) of the microscope supporting arm.
                  Auto. Mode:      Run the automatic sequence of photo collection.

    AUTHOR : Danilo Roccatano (c)
    VERSION: 1.0
    DATE   : August 2021
    
*/

// include Arduino stepper motor library
#include <Servo.h>
#include <Stepper.h>
// include the library code for the LCD
#include <LiquidCrystal.h>

//LCD settings
// Set the LCD address to 0x27 for a 16 chars and 2 line display
const uint8_t I2C_ADDRESS = 0x27; //watch video. use I2C Scanenr to get the address
const uint8_t LCD_CHAR = 16;
const uint8_t LCD_LINE = 2;

char MENU_TITLE[3][20] = {"X-steps/degree:", "Y-steps/degrees:", "Auto. Mode"};

//  Stepper setup variables
// define total number of steps per revolution
#define STEPS 2038
// Define the maximum number of step for the Y-stepper (~100 degrees)
#define MAXSTEP 1400

// push button pin
#define PBUTTON_PIN 31

// define USB microscope stepper motor control pins
#define IN1  28
#define IN2  26
#define IN3  24
#define IN4  22

// define rotation table  stepper motor control pins
#define IN21   36
#define IN22   34
#define IN23   32
#define IN24   30

// Joystick module setup
#define GROUND_JOY_PIN A8              //joystick ground pin will connect to Arduino analog pin A8
#define VOUT_JOY_PIN A9                //joystick +5 V pin will connect to Arduino analog pin A9
#define YJOY_PIN A10                   //X axis reading from joystick will go into analog pin A10
#define XJOY_PIN A11                   //Y axis reading from joystick will go into analog pin A11
#define BUTTON_PIN 39                  //Y axis reading from joystick will go into analog pin 


// Redefine pins for extra VCC and Ground

#define VCC2  33 // define  digital pin here as VCC
#define GND2  35 // define  digital pin as Ground 
#define VCC3  39 // define  digital pin here as VCC
#define GND3  41 // digital pin as Ground 

int buttonState = 0;
int nsp1 = 100;
int nsp2 = 100;
int nspd1 ;
int nspd2 ;
int nsp = 100;
int speed1 = 5;
int speed2 = 5;
int nystep = 0; // current number of y step
int maxystep = 0;
int srun = -1;
int sensor = 0;
int nst = 0;
int imenu = 0;          // select menu item

int buttonPin = 2;

// setup the servo object name for the blade movement
Servo myservo ; // slice bar control

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);

// initialize stepper library
Stepper stepper(STEPS, IN4, IN2, IN3, IN1);
// Initialize the stepper for the object table
Stepper stepper1(STEPS, IN24, IN22, IN23, IN21);

void setup() {
  Serial.begin(9600);
  // Nothing to do (Stepper Library sets pins as outputs)

  pinMode(PBUTTON_PIN, INPUT);
  // set up the LCD's number of columns and rows:
  lcd.begin(16, 2);

  // Define the pin mode for powering the
  pinMode(VCC2, OUTPUT); //define a digital pin as output
  digitalWrite(VCC2, HIGH);// set the above pin as HIGH so it acts as 5V
  pinMode(VCC3, OUTPUT); //define a digital pin as output
  digitalWrite(VCC3, HIGH);// set the above pin as HIGH so it acts as 5V

  pinMode(GND2, OUTPUT); //define a digital pin as output
  digitalWrite(GND2, LOW);// set the above pin as LOW so it acts as Ground
  pinMode(GND3, OUTPUT); //define a digital pin as output
  digitalWrite(GND3, LOW);// set the above pin as LOW so it acts as Ground

  pinMode(VOUT_JOY_PIN, OUTPUT) ;    //pin A3 shall be used as output
  pinMode(GROUND_JOY_PIN, OUTPUT) ;  //pin A2 shall be used as output
  digitalWrite(VOUT_JOY_PIN, HIGH) ; //set pin A3 to high (+5V)
  digitalWrite(GROUND_JOY_PIN, LOW) ; //set pin A3 to low (ground)
  pinMode(buttonPin, INPUT_PULLUP);

  myservo.attach(2);
}

void loop() {

  // home the microscope at the starting
  if (srun == -1) {
    HomePos();
    // Print a welcome message to the LCD.
    lcd.setCursor(0, 0);
    lcd.print("MicroScanner V.1");
    lcd.setCursor(0, 1);
    lcd.print("Use Joystick/Key");
    delay(1000);
    DisplayMenu(imenu);
  }

  int x = analogRead (0);

/*
 * KEY PAD CONTROL
 */

  // Right key
  if (x < 60) {
    DisplayMenu(imenu);
    switch (imenu) {
      case 0:
        nspd1++;
        if (nspd1 > 30) nspd1 = 1;
        nsp1 = 1176 * nspd1 / 90.;
        clearLine2();
        lcd.setCursor(0, 1);
        lcd.print("S:");
        lcd.print(nsp1);
        lcd.print(" A: ");
        lcd.print(nspd1);
        break;
      case 1:
        nspd2++;
        if (nspd2 > 30) nspd2 = 1;
        nsp2 = 1176 * nspd2 / 90.;
        clearLine2();
        lcd.setCursor(0, 1);
        lcd.print("S:");
        lcd.print(nsp2);
        lcd.print(" A: ");
        lcd.print(nspd2);
        break;
      case 2:
        lcd.setCursor(0, 0);
        lcd.print("MicroScanner V.1");
        HomePos();
        AutoRun();
    }
    delay(200);
  }
  
  // MENU control:Up key
  else if (x < 200) {
    imenu++;
    if (imenu > 2) imenu = 0;
    DisplayMenu(imenu);

  }
  
  // MENU control: Down key
  else if (x < 400) {
    imenu--;
    if (imenu < 0) imenu = 2;
    DisplayMenu(imenu);
  }
  
  // Left key
  else if (x < 600) {
    DisplayMenu(imenu);
    switch (imenu) {
      case 0:
        nspd1--;
        if (nspd1 < 1) nspd1 = 30;
        nsp1 = 4703.*nspd1 / 360.;
        clearLine2();
        lcd.setCursor(0, 1);
        lcd.print("S:");
        lcd.print(nsp1);
        lcd.print(" A: ");
        lcd.print(nspd1);
        break;
      case 1:
        nspd2--;
        if (nspd2 < 1) nspd2 = 30;
        nsp2 = 4703.*nspd2 / 360.;
        clearLine2();
        lcd.setCursor(0, 1);
        lcd.print("S:");
        lcd.print(nsp2);
        lcd.print(" A: ");
        lcd.print(nspd2);
        break;
    }
    delay(200);

  }
  else if (x < 800) {
    //select button is used for homing the microscope inclination
    clearLine2();
    lcd.print ("Home");
    HomePos();
  }
  
  /* 
   *  Check if the the Y movement reach the home position
   */
   
  buttonState = digitalRead(PBUTTON_PIN);
  if (!buttonState) {
    nystep = 0;
    maxystep = 0;
  }
  
/*
 *  JOYSTICK MODULE CONTROL
 */
  // read analog value from the X potentiometer

  int joystickXVal = analogRead(XJOY_PIN) ;
  Serial.print("X-STICK:" );
  Serial.println(joystickXVal);

  if (  joystickXVal <  450 )
  {
    clearLine1();
    lcd.setCursor(0, 0);
    lcd.print ("Rotate Platform  ");
    nsp=nsp2;
    RotateT();
  } else if  (joystickXVal >  550) {
    clearLine1();
    lcd.setCursor(0, 0);
    lcd.print ("Rotate Platform  ");
    nsp = -nsp2;
    RotateT();
  }

  int joystickYVal = analogRead(YJOY_PIN) ;
  Serial.print("Y-STICK:" );
  Serial.println(joystickYVal);

  if (  joystickYVal < 450 )
  {
    clearLine1();
    lcd.setCursor(0, 0);
    lcd.print ("Move Up  ");
    nsp = nsp1;
    MoveUp();
  } else if (joystickYVal > 550 ) {
    clearLine1();
    lcd.setCursor(0, 0);
    lcd.print ("Move Down  ");
    nsp = -nsp1;
    MoveUp();
  }

}

/*
 *******************FUNCTIONS
 */
 
void DisplayMenu( int imenu) {

  lcdDisplayLine1(imenu);
  switch (imenu) {
    case 0:
      lcdDisplayLine2(nsp1);
      break;
    case 1:
      lcdDisplayLine2(nsp2);
      break;
    case 2:
     lcd.setCursor(0, 1);
     lcd.print("-> Key to Start");
      break;
    default:
      clearLine2();
  }
}

void lcdDisplayLine1(int imenu)
{
  clearLine1();
  lcd.setCursor(0, 0);
  lcd.print(MENU_TITLE[imenu]);

} //

void lcdDisplayLine2(int value)
{
  clearLine2();
  lcd.setCursor(0, 1);
  lcd.print(value);//print value
}

void AutoRun() {

  // Run the sequence for data collection

  // Calculate the number of rotation step for the table (nrsx) and microscope (nrsy).

  int nrsy = MAXSTEP / nsp1;
  int nrsx = STEPS / nsp2;

  lcd.setCursor(0, 1);
  lcd.print("Select Key:Abort");

  // Rotate the microscope
  for (int ii = 1; ii <= nrsy; ii++) {
    stepper.setSpeed(speed1);
    stepper.step(nsp1);

    // Rotate the sample table
    for (int kk = 1; kk <= nrsx; kk++) {

      // Read the joystick button state
      buttonState = digitalRead(buttonPin);
      delay(100);

      // Abort procedure when the select key button is pushed
      int x = analogRead (0);
      if (x > 600 && x < 800) return;

      stepper1.setSpeed(speed2);
      stepper1.step(nsp2);

      // Take a Picture
      ShutterControl();

    }
  }
}

void  RotateT() {
  Serial.print("speed2: ");
  Serial.println(speed2);
  Serial.print("nsp2: ");
  Serial.println(nsp);
  // Rotate the sample table
  stepper1.setSpeed(speed2);
  stepper1.step(nsp);
}

void MoveUp()
{
  // Move microscope's inclination stepper

  if (nystep + nsp <= MAXSTEP) {
    stepper.setSpeed(speed1);
    stepper.step(nsp);
    nystep += nsp;
  } else {
    maxystep = nystep;
  }

  printYSteps(nystep);
}

void printYSteps(int value) {

  float angle;
  angle = value * 360. / 4703.;

  clearLine2();
  lcd.setCursor(0, 1);
  lcd.print("S:");
  lcd.print(value);
  lcd.print(" A: ");
  lcd.print(angle);
}

void HomePos()
{
   lcd.setCursor(0, 0);
  lcd.print("MicroScanner V.1");
  lcd.setCursor(0, 1);
  lcd.print("Homing...       ");
  nsp = -50;
  // start to rotate the y motor
  nystep = 0;
  srun = 0;
  buttonState = digitalRead(PBUTTON_PIN);
  while (buttonState) {
    stepper.setSpeed(5);
    stepper.step(nsp);
    buttonState = digitalRead(PBUTTON_PIN);
  }
  clearLine2();
}

void ShutterControl() {

  // Press the shutter for the photo
  delay(500);
  myservo.write(25);
  delay(500);
  myservo.write(75);
  delay(1000);

}

void clearLine1()
{
  for (int i = 0; i <= LCD_CHAR - 1; i++)
  {
    lcd.setCursor (i, 0);
    lcd.write(254);
  }
}

void clearLine2()
{
  for (int i = 0; i <= LCD_CHAR - 1; i++)
  {
    lcd.setCursor (i, 1); //
    lcd.write(254);
  }
}
